from doka.kernel.repository.repositories import UnitRepository, ProjectRepository, StackRepository
from doka.kernel.definition.entities import UnitSchema, ProjectSchema, StackSchema
from doka.kernel.services.git import Git
import json


class Inspect(object):
    def __init__(self, project=None, filter=None, stack=None, name=None, full: bool = None):
        self._flags = {
            "project": project,
            "stack": stack,
            "name": name,
            "full": full
        }
        try:
            self._flags["filter"] = json.loads(filter)
        except:
            self._flags["filter"] = {}

        if project:
            self._project = ProjectRepository().get_one(project)
        else:
            self._project = ProjectRepository().get_active()

    def projects(self):
        projects = ProjectRepository().get()
        return self._represent(projects, ProjectSchema, ["name", "env", "active"])

    def stacks(self):
        stacks = StackRepository(project=self._project).get()
        return self._represent(stacks, StackSchema, ["name", "network"])

    def units(self):
        search_filter = self._flags.get("filter")
        if self._flags.get("stack"):
            search_filter["stack_name"] = self._flags.get("stack")
        if self._flags.get("project"):
            search_filter["project_name"] = self._flags.get("project")
        if self._flags.get("name"):
            search_filter["name"] = self._flags.get("stack")

        units = UnitRepository(self._project).get(search_filter or None)
        for unit in units:
            if unit.repo:
                unit.branch = Git(unit).get_branch()

        return self._represent(units, UnitSchema, ["name", "branch", "stack_name"])

    def _represent(self, collection, schema, brief: list):
        result = []
        for entity in schema(many=True).dump(collection).data:
            item = {}
            for k, v in entity.items():
                if v is None:
                    continue
                if not self._flags.get("full") and k not in brief:
                    continue
                item[k] = v
            result.append(item)
        return result



import subprocess
from doka.kernel.definition.entities import UnitEntity, ProjectEntity, StackEntity
from doka.kernel.storage.yaml import YamlStorage
from doka.kernel.repository.repositories import UnitRepository, StackRepository
from doka.kernel.storage.hooks import move_unit_environment_to_stack, move_project_environment_to_stack


class Runtime:
    def get_unit_ip(self, unit_entity: UnitEntity):
        return subprocess.check_output(
            "docker inspect -f '{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}' " + "_".join([
                unit_entity.project_name,
                unit_entity.container_name,
                '1'
            ]),
            shell=True).decode().strip()


class Environment:
    def set_for_stack(self, stack: StackEntity):
        units_repository = UnitRepository(stack.project_name)
        for unit in units_repository.get({"stack_name": stack.name, "repo": {"$neq": None}}):
            self.set_for_unit(unit)

    def set_for_project(self, project: ProjectEntity):
        move_project_environment_to_stack(project)

    @staticmethod
    def set_for_unit(unit: UnitEntity):
        move_unit_environment_to_stack(unit)


class Volume:
    def create(self, stack: StackEntity):
        if not stack.volumes:
            return
        for name, settings in stack.volumes.items():
            if settings is not None and settings.get("external"):
                subprocess.run(
                    "docker volume create {name} || true".format(
                        name=name
                    ),
                    shell=True,
                    check=True
                )


class Network:
    def create(self, stack: StackEntity):
        subprocess.run(
            "docker network create {name} || true".format(
                name="_".join([stack.project_name, stack.network])
            ),
            shell=True,
            check=True
        )

    def join(self, stack: StackEntity, network_name: str):
        compose_file = stack.compose_directory.joinpath(stack.compose_file)
        storage = YamlStorage(compose_file)
        storage.append("networks", {network_name: {"external": True}})
        storage.append("services.*.networks", network_name)


class Unit:
    def __init__(self, unit: UnitEntity):
        self._unit = unit
        self._stack = StackRepository(self._unit.project_name).get_one(unit.stack_name)

    def build(self, no_cache=None):
        subprocess.run(
            "docker build -t {name} {no_cache} {directory}".format(
                name=self._unit.image_name,
                directory=self._unit.codedir,
                no_cache='--no-cache' if no_cache else ''
            ),
            shell=True,
            check=True
        )

    def start(self):
        subprocess.run(
            "cd {directory} && docker-compose -p {project_name} -f {compose_file} up -d {container_name}".format(
                project_name=self._unit.project_name,
                compose_file=self._stack.compose_file,
                directory=self._stack.compose_directory,
                container_name=self._unit.container_name
            ),
            shell=True,
            check=True
        )

    def stop(self):
        subprocess.run(
            "cd {directory} && docker-compose -p {project_name} -f {compose_file} stop {container_name} && "
            "docker-compose -p {project_name} -f {compose_file} rm -f {container_name}".format(
                project_name=self._unit.project_name,
                compose_file=self._stack.compose_file,
                directory=self._stack.compose_directory,
                container_name=self._unit.container_name
            ),
            shell=True,
            check=True
        )

    def restart(self):
        subprocess.run(
            "cd {directory} && docker-compose -p {project_name} -f {compose_file} restart {container_name}".format(
                project_name=self._unit.project_name,
                compose_file=self._stack.compose_file,
                directory=self._stack.compose_directory,
                container_name=self._unit.container_name
            ),
            shell=True,
            check=True
        )

    def scale(self, num):
        subprocess.run(
            "cd {directory} && docker-compose -p {project_name} -f {compose_file} scale {container_name}={num}".format(
                project_name=self._unit.project_name,
                compose_file=self._stack.compose_file,
                directory=self._stack.compose_directory,
                container_name=self._unit.container_name,
                num=num
            ),
            shell=True,
            check=True
        )


class Stack:
    def __init__(self, stack: StackEntity):
        self._stack = stack

    def up(self):
        subprocess.run(
            "cd {directory} && docker-compose -p {project_name} -f {compose_file} up -d".format(
                project_name=self._stack.project_name,
                compose_file=self._stack.compose_file,
                directory=self._stack.compose_directory
            ),
            shell=True,
            check=True
        )

    def down(self):
        subprocess.run(
            "cd {directory} && docker-compose -p {project_name} -f {compose_file} down".format(
                project_name=self._stack.project_name,
                compose_file=self._stack.compose_file,
                directory=self._stack.compose_directory
            ),
            shell=True,
            check=True
        )



from doka.commands.unit import Unit
from doka.commands.project import Project
from doka.commands.stack import Stack
from doka.commands.inspect import Inspect
from doka.commands.portainer import Portainer
from doka.commands.config import Config
from doka.commands.nginx import Nginx



import subprocess


class Portainer(object):
    def up(self):
        subprocess.run(
            "docker volume create portainer_data || true",
            shell=True,
            check=True,
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL
        )
        subprocess.run(
            "if docker inspect portainer | grep exited > /dev/null; then docker restart portainer; fi",
            shell=True,
            check=True,
            stderr=subprocess.DEVNULL
        ),
        subprocess.run(
            "if "
            "docker inspect -f '{{.State.Running}}' portainer | grep true > /dev/null; then echo \"Nothing to do, portainer is running\"; "
            "else "
            "docker run -d --name portainer -p 9100:9000 -v /var/run/docker.sock:/var/run/docker.sock -v portainer_data:/data portainer/portainer && echo \"Portainer has started\"; "
            "fi",
            shell=True,
            check=True,
            stderr=subprocess.DEVNULL
        )

    def down(self):
        subprocess.run(
            "docker stop portainer && docker rm portainer",
            shell=True,
            check=True
        )



#!/usr/bin/env python3
import subprocess
from doka.kernel.definition.entities import UnitEntity
import shutil


class Git:
    def __init__(self, unit: UnitEntity):
        self._unit = unit

    def clone(self, from_template=False):
        subprocess.run(
            "rm -rf {directory} && git clone {repo} {directory}/".format(
                repo=self._unit.repo,
                directory=self._unit.codedir
            ),
            shell=True,
            check=True)
        if from_template:
            shutil.rmtree('{}/.git'.format(self._unit.codedir))

    def init(self):
        subprocess.run(
            "cd {directory} && git init && git add readme.md && git commit -m 'Initial commit' ".format(
                directory=self._unit.codedir),
            shell=True,
            check=True
        )

    def add_origin(self, push_initial_commit=False):
        subprocess.run(
            "cd {directory} && git remote add origin {repo} {initial_commit}".format(
                directory=self._unit.codedir, repo=self._unit.repo,
                initial_commit='&& git push origin master' if push_initial_commit else ''
            ),
            shell=True,
            check=True
        )

    def checkout(self, branch, force=False):
        subprocess.run("cd {directory} && git checkout {force} {branch}".format(
            directory=self._unit.codedir,
            force='-b' if force else '',
            branch=branch
        ), shell=True, check=True)

    def fetch(self, force=True):
        subprocess.run("cd {directory} && git fetch {force}".format(
            directory=self._unit.codedir,
            force='&& git reset --hard origin/development' if force else ''
        ), shell=True, check=True)

    def pull(self):
        subprocess.run("cd {directory} && git pull origin {branch}".format(
            directory=self._unit.codedir,
            branch=self.get_branch()
        ), shell=True, check=True)

    def get_branch(self):
        return subprocess.check_output(
            "cd {directory} && git rev-parse --abbrev-ref HEAD".format(directory=self._unit.codedir),
            shell=True).decode().strip()

    def commit(self, message=''):
        subprocess.run("cd {directory} && git add . && git commit -m \"{message}\"".format(
            directory=self._unit.codedir,
            message=message
        ), shell=True, check=True)

    def merge(self, dst_branch):
        subprocess.run("cd {directory} && git checkout {dst_branch} && git merge {src_branch}".format(
            directory=self._unit.codedir,
            dst_branch=dst_branch,
            src_branch=self.get_branch()
        ), shell=True, check=True)

    def push(self):
        subprocess.run("cd {directory} && git push origin {branch}".format(
            directory=self._unit.codedir,
            branch=self.get_branch()
        ), shell=True, check=True)



def entity_filter(filter):
    def filter_function(entity):
        def calculate_operation(a, operation, b):
            if operation == '$eq':
                return a == b
            elif operation == '$neq':
                return a != b

        for field, condition in filter.items():
            if not isinstance(condition, dict):
                condition = {"$eq": condition}

            fl = True
            for operation, value in condition.items():
                if not calculate_operation(getattr(entity, field), operation, value):
                    fl = False
                    break
            if not fl:
                return False

        return True

    return filter_function



from doka.kernel.definition.entities import UnitEntity
from doka.kernel.services.git import Git
import doka.kernel.services.docker as docker


def build(unit_entity: UnitEntity):
    docker.Unit(unit_entity).build()


def checkout(unit_entity: UnitEntity, branch, is_build=True):
    Git(unit_entity).checkout(branch)
    if is_build:
        build(unit_entity)


def pull(unit_entity: UnitEntity, is_build=True):
    Git(unit_entity).pull()
    if is_build:
        build(unit_entity)


def fetch(unit_entity: UnitEntity):
    Git(unit_entity).fetch()


def start(unit_entity: UnitEntity):
    docker.Unit(unit_entity).start()


def stop(unit_entity: UnitEntity):
    docker.Unit(unit_entity).stop()


def restart(unit_entity: UnitEntity):
    docker.Unit(unit_entity).restart()


def scale(unit_entity: UnitEntity, num):
    docker.Unit(unit_entity).scale(num)



from doka.kernel.services import docker
from doka.kernel.definition.entities import StackEntity, ProjectEntity, UnitEntity
from doka.kernel.repository.repositories import UnitRepository
from doka.kernel.actions import unit
from doka.kernel.storage.yaml import YamlStorage
from doka.kernel.storage import hooks


def join_network(stack_entity: StackEntity, network_name: str):
    docker.Network().join(stack_entity, network_name)


def register_unit_in_compose_file(stack_entity: StackEntity, unit_entity: UnitEntity):
    storage = YamlStorage(stack_entity.compose_directory.joinpath(stack_entity.compose_file))
    storage.set(
        index='services',
        value={
            unit_entity.container_name: {
                'image': unit_entity.name,
                'env_file': '.{}.env'.format(unit_entity.name),
                'networks': ["_".join([stack_entity.project_name, stack_entity.network])]
            }
        }
    )
    storage.write()


def unregister_unit_in_compose_file(stack_entity: StackEntity, unit_entity: UnitEntity):
    storage = YamlStorage(stack_entity.compose_directory.joinpath(stack_entity.compose_file))
    storage.remove(
        index='services.{}'.format(unit_entity.container_name)
    )
    storage.write()


def deploy(project_entity: ProjectEntity, stack_entity: StackEntity):
    storage = YamlStorage(stack_entity.compose_directory.joinpath(stack_entity.compose_file))
    for unit_entity in UnitRepository(stack_entity.project_name).get_for_build(stack_entity.name):
        storage.set(
            index='services.{}.image'.format(unit_entity.container_name),
            value="{}/{}".format(project_entity.docker_registry, unit_entity.repo.split(":").pop().replace('.git', ''))
        )
        hooks.move_unit_environment_to_deploy(project_entity, unit_entity)
    storage.write(project_entity.metadir.joinpath('deploy', stack_entity.name).with_suffix('.yml'))


def build(stack_entity: StackEntity):
    for unit_entity in UnitRepository(stack_entity.project_name).get_for_build(stack_entity.name):
        try:
            unit.build(unit_entity)
        except Exception as e:
            print("!!!Cannot build unit {}. Error: ".format(unit_entity.name), str(e))


def up(stack_entity: StackEntity, project_entity: ProjectEntity):
    storage = YamlStorage(stack_entity.compose_directory.joinpath(stack_entity.compose_file))
    if not storage.read().get("services"):
        return
    environment = docker.Environment()
    environment.set_for_project(project_entity)
    environment.set_for_stack(stack_entity)
    docker.Network().create(stack_entity)
    docker.Volume().create(stack_entity)
    docker.Stack(stack_entity).up()


def down(stack_entity: StackEntity):
    storage = YamlStorage(stack_entity.compose_directory.joinpath(stack_entity.compose_file))
    if not storage.read().get("services"):
        return
    docker.Stack(stack_entity).down()


def checkout(stack_entity: StackEntity, branch: str):
    for unit_entity in UnitRepository(stack_entity.project_name).get_for_build(stack_entity.name):
        try:
            unit.checkout(unit_entity, branch, is_build=False)
        except Exception:
            print("!!!Cannot checkout repository: ", unit_entity.name)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['doka',
 'doka.commands',
 'doka.kernel',
 'doka.kernel.actions',
 'doka.kernel.definition',
 'doka.kernel.repository',
 'doka.kernel.services',
 'doka.kernel.storage',
 'doka.plugins',
 'doka.plugins.nginx']

package_data = \
{'': ['*'],
 'doka': ['tests/*'],
 'doka.plugins.nginx': ['conf/*', 'templates/*']}

install_requires = \
['fire>=0.1.3,<0.2.0',
 'marshmallow>=2.19,<3.0',
 'python-dotenv>=0.10.3,<0.11.0',
 'pyyaml>=5.1,<6.0',
 'requests>=2.22,<3.0',
 'termcolor>=1.1,<2.0']

entry_points = \
{'console_scripts': ['doka = doka:main']}

setup_kwargs = {
    'name': 'doka',
    'version': '0.4.4',
    'description': '',
    'long_description': None,
    'author': 'axxil',
    'author_email': 'axxil@yandex.ru',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from doka.kernel.definition.entities import UnitEntity, StackEntity, ProjectEntity, StackSchema, UnitSchema


def represent_none(self, _):
    return self.represent_scalar('tag:yaml.org,2002:null', '')


def _extract_image_name(repo):
    return repo.split("/").pop().replace('.git', '').split('.').pop().replace('-', '_')


def create_unstacked_stack(project_entity: ProjectEntity):
    from doka.kernel.repository.repositories import StackRepository
    import yaml

    stack_entity = StackEntity(
        name='unstacked',
        project_name=project_entity.name,
        compose_directory=project_entity.metadir.joinpath("stacks"),
        network="default"
    )

    with project_entity.metadir.joinpath("stacks", stack_entity.name).with_suffix('.yml').open('wt') as f:
        yaml.dump(
            {
                "version": '2',
                "services": None,
                "networks": {
                    "_".join([project_entity.name, stack_entity.network]): {
                        "external": True
                    }
                }
            },
            f,
            default_flow_style=False,
            sort_keys=False
        )

        StackRepository(project_entity).save(stack_entity)


def project_post_sync(project_name):
    import json
    import yaml
    import shutil
    from doka.kernel.repository.repositories import ProjectRepository

    project_entity = ProjectRepository().get_one(project_name)

    yaml.add_representer(type(None), represent_none)

    stacks = []
    units = []
    for entry in project_entity.metadir.joinpath("stacks").glob('**/*.yml'):
        with entry.open('rt') as f:
            yml = yaml.load(f, Loader=yaml.FullLoader)

        stack = StackEntity(
            name=entry.name.replace('.yml', ''),
            project_name=project_name,
            compose_directory=project_entity.metadir.joinpath("stacks"),
            network="default",
            volumes=yml.get("volumes")
        )

        services = yml.get("services")
        for unit_name, unit_settings in services.items():
            unit_entity = UnitEntity(
                name="{}.{}".format(stack.name, unit_name),
                image_name=None,
                project_name=project_name,
                stack_name=stack.name
            )
            unit_settings["networks"] = ["_".join([project_name, stack.network])]
            if "build" in unit_settings:
                image_name = _extract_image_name(unit_settings.get("build"))
                unit_entity.image_name = "_".join([project_name, stack.name, image_name])
                unit_entity.codedir = project_entity.workdir.joinpath(*[stack.name, image_name])
                unit_entity.repo = unit_settings.get("build")
                unit_entity.working_dir = unit_settings.get("working_dir") or '/code'

                unit_settings["image"] = unit_entity.image_name
                unit_settings["env_file"] = ".{}.env".format(unit_entity.name)

                unit_settings.pop("build", None)
                unit_settings.pop("working_dir", None)

            units.append(unit_entity)

        yml["services"] = services

        yml["networks"] = {
            "_".join([project_name, stack.network]): {
                "external": True
            }
        }

        with entry.open("wt") as f:
            yaml.dump(yml, f, default_flow_style=False, sort_keys=False)

        stacks.append(stack)

    with open(project_entity.metadir.joinpath("units.json"), 'wt') as f:
        json.dump(UnitSchema(many=True).dump(units).data, f, indent=2)

    with open(project_entity.metadir.joinpath("stacks.json"), "wt") as f:
        json.dump(StackSchema(many=True).dump(stacks).data, f, indent=2)

    shutil.rmtree(project_entity.metadir.joinpath(".git"))


def move_project_environment_to_stack(project: ProjectEntity):
    import shutil
    shutil.copyfile(
        src=project.metadir.joinpath("environments", ".env.{}".format(project.env)),
        dst=project.metadir.joinpath("stacks", ".env".format(project.env)),
    )


def move_unit_environment_to_stack(unit_entity: UnitEntity):
    from doka.kernel.storage.env import ProjectEnvStorage, UnitEnvStorage, StacksEnvStorage
    from doka.kernel.repository.repositories import ProjectRepository

    project_environment = ProjectEnvStorage(ProjectRepository().get_one(unit_entity.project_name)).read()

    try:
        unit_environment = UnitEnvStorage(unit_entity).read() or {}
    except FileNotFoundError:
        unit_environment = {}
        print("!!!Cannot find .env file for unit {}. Empty environment set".format(unit_entity.name))

    env = {}
    for k, v in unit_environment.items():
        if not v:
            v = project_environment.get(k, v)
        env[k] = v
    StacksEnvStorage(unit_entity).save(env)


def move_unit_environment_to_deploy(project_entity: ProjectEntity, unit_entity: UnitEntity):
    import shutil
    src = unit_entity.codedir.joinpath('.env.example')
    dst = project_entity.metadir.joinpath('deploy', ".{}.env".format(unit_entity.name))
    shutil.copy(src, dst)


def delete_unit_environment_from_stack(unit_entity: UnitEntity):
    from doka.kernel.repository.repositories import ProjectRepository
    project_entity = ProjectRepository().get_one(unit_entity.project_name)
    project_entity.metadir.joinpath(*['stacks', '.{}.env'.format(unit_entity.name)]).unlink()
